#ifndef MSP_GL_TEXTURE2D_H_
#define MSP_GL_TEXTURE2D_H_

#include <string>
#include <msp/graphics/image.h>
#include "texture.h"

namespace Msp {
namespace GL {

/**
Two-dimensional texture.  Consists of an array of texels in the shape of a
rectangle.  Texture coordinate have a range of [0, 1].  Coordinates outside of
this range are subject to wrapping.  This is the most common type of texture.
*/
class Texture2D: public Texture
{
public:
	class Loader: public Msp::DataFile::DerivedObjectLoader<Texture2D, Texture::Loader>
	{
	public:
		Loader(Texture2D &);
		Loader(Texture2D &, Collection &);
	private:
		void init();

		void raw_data(const std::string &);
		void storage(PixelFormat, unsigned, unsigned);
		void storage_levels(PixelFormat, unsigned, unsigned, unsigned);
	};

private:
	class AsyncLoader;

	unsigned width;
	unsigned height;
	unsigned levels;
	unsigned allocated;

public:
	Texture2D(ResourceManager * = 0);
	virtual ~Texture2D();

	/** Defines storage structure for the texture.  If lv is zero, the number
	of mipmap levels is automatically determined from storage dimensions.

	Must be called before an image can be uploaded.  Once storage is defined,
	it can't be changed. */
	void storage(PixelFormat fmt, unsigned wd, unsigned ht, unsigned lv = 0);

	DEPRECATED void storage(PixelComponents cm, unsigned wd, unsigned ht, unsigned lv = 0)
	{ storage(make_pixelformat(cm, UNSIGNED_BYTE), wd, ht, lv); }

	/** Allocates storage for the texture.  The contents are initially
	undefined.  If storage has already been allocated, does nothing. */
	void allocate(unsigned level);

	/** Updates the contents of the entire texture.  Storage must be defined
	beforehand.  The image data must have dimensions and format matching the
	defined storage. */
	virtual void image(unsigned level, const void *data);

	DEPRECATED void image(unsigned level, PixelComponents fmt, DataType type, const void *data);

	/** Updates a rectangular region of the texture.  Storage must be defined
	beforehand.  The image data must be in a format mathing the defined storage
	and the update region must be fully inside the texture. */
	void sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht, const void *data);

	DEPRECATED void sub_image(unsigned level, int x, int y, unsigned wd, unsigned ht,
		PixelComponents fmt, DataType type, const void *data);

	/** Updates the contents of the entire texture from an image.  If storage
	has not been defined, it will be set to match the image.  Otherwise the
	image must match the defined storage. */
	virtual void image(const Graphics::Image &, unsigned lv = 0);

	using Texture::image;

private:
	void image(const Graphics::Image &, unsigned, bool);

public:
	unsigned get_width() const { return width; }
	unsigned get_height() const { return height; }

private:
	unsigned get_n_levels() const;
	void get_level_size(unsigned, unsigned &, unsigned &) const;

public:
	virtual Resource::AsyncLoader *load(IO::Seekable &, const Resources * = 0);
	virtual UInt64 get_data_size() const;
	virtual void unload();
};

} // namespace GL
} // namespace Msp

#endif
